// ==UserScript==
// @name         Equalizer_IMAC_Desktop.user
// @namespace    maid.aural.inspection
// @version      1
// @description  Firefox137向け：Equalizer_IMAC_Desktop.user
// @author       Gemini
// @match        https://*youtube.com/*
// @grant        none
// ==/UserScript==

(async function() {
    'use strict';
    

    // --- 設定変数 ---
    const TAP_AUDIO_DURATION_SEC = 3.00; // 基本周波数推定のためのデータ収集時間（秒）
    const TARGET_SAMPLE_RATE = 96000; // 目標サンプルレート (Hz)

    // メディア要素待機設定
    const MEDIA_WAIT_INTERVAL_MS = 300;
    const MEDIA_WAIT_MAX_TIME_SEC = 20;

    // LocalEnhance 処理用設定 (Worklet内で使用)
    const LE_CHUNK_SIZE_WORKLET = 128;
    const LE_RMS_THRESHOLD_WORKLET = 0.008; // RMS閾値 (さらに敏感に)

    // 「弾性波動」関連定数 (Worklet内で使用)
    const EW_VELOCITY = 340.29;
    const EW_WAVELENGTH_FACTOR = EW_VELOCITY / 440;
    const EW_ENHANCEMENT_FACTOR = Math.sin(Math.PI / EW_WAVELENGTH_FACTOR) * 1.0; // 効果をさらに強調

    // 「Gamma式補完」関連定数 (Worklet内で使用 - 再設計版、補間強化、概念応用)
    const GAMMA_L = 1.00;
    const GAMMA_T = 1.00;
    const GAMMA_MU0 = 1.00;
    const GAMMA_KAPPA = 1.00;
    const GAMMA_ALPHAP = 1.00;
    const GAMMA_THRESHOLD_WORKLET = 0.2; // Gamma閾値をさらに下げて効果を頻繁に
    const GAMMA_INTERPOLATION_FACTOR = 0.8; // 補間値をどの程度混ぜるか (0.0から1.0)
     const GAMMA_DYNAMIC_GAIN_MULTIPLIER = 1.5; // 動的ゲイン調整の最大乗数

    // グローバー反復回数概念応用設定 (Worklet内で使用)
    const GROVER_ITERATION_BASE = 2; // 反復回数のベース（概念的なもの）
    const GROVER_OPTIMIZATION_FACTOR = 0.5; // 信号特徴による最適化の度合い

    // 小数点処理設定 (Worklet内で使用)
    const FIXED_PRECISION_DECIMALS = 2; // 小数点以下第2位

    // --- 音響物理学応用設定 ---
    const RHO0 = 1.2;
    const C = 343;
    const HORN_L = 0.5;
    const HORN_M = 2.5;
    const HORN_S0 = 0.01;
    const HORN_CUTOFF_FREQ_CALCULATED = HORN_M * C / (4 * Math.PI);
    const TUBE_LENGTH = 0.5;
    const TUBE_RESONANCE_FREQ = (1 * C) / (2 * TUBE_LENGTH);


     // --- IMAX再現強化設定 (v8.0ベースから維持) ---

     // サブハーモニック合成設定
    const SUBHARMONIC_RATIO = 2;
    const SUBHARMONIC_GAIN = 0.7; // さらに強調

     // Mid/Side 処理設定
    const MID_GAIN_DB = -3; // さらに抑える
    const SIDE_GAIN_DB = 8; // 最大限に強調

     // 簡易空間反射模倣設定 (Convolutionを主とするため、簡易反射は補助)
     const EARLY_REFLECTION_DELAYS_MS = [4, 8, 12]; // ディレイを短めに、数を減らす
     const EARLY_REFLECTION_GAINS = [0.15, 0.1, 0.08]; // ゲインを控えめに

     // Convolution Reverb 設定
     const CONVOLVER_IR_DURATION_SEC = 0.3; // IRを少し長く
     const CONVOLVER_IR_DECAY_RATE = 4.0; // 減衰を少し緩やかに
     const CONVOLVER_IR_NUM_REFLECTIONS = 7; // 反射数を増やす
     const CONVOLVER_WET_GAIN = 0.5; // Convolution後の信号のミックス量
     const CONVOLVER_DRY_GAIN = 0.8; // ドライ信号のミックス量


     // EQ設定調整 - さらにIMAX強調
     const ULTIMATE_IMAX_EQ_SETTINGS_ADVANCED = [
        { type: 'lowshelf', freq: 18, gain: 18 }, // 超低域を極限まで強調
        { type: 'peaking', freq: 40, gain: 8, Q: 2.5 }, // 低域のパンチを極限まで強化
        { type: 'peaking', freq: 80, gain: 4, Q: 2.0 }, // 低域の量感をさらに強調
        { type: 'peaking', freq: 250, gain: -4, Q: 1.0 }, // 低中域のブーミーさをさらに抑制
        { type: 'peaking', freq: 2000, gain: 4, Q: 4.0 }, // 中高域のプレゼンスを極限まで強調
        { type: 'highshelf', freq: 17000, gain: 6 }, // 高域のエア感を極限まで強調
         { type: 'highshelf', freq: 20000, gain: -8 }, // 超高域を非常に強くロールオフ
    ];


     // ダイナミクス設定調整 - さらにIMAX強調
     const ULTIMATE_IMAX_COMPRESSOR_THRESHOLD_ADVANCED = -8; // dB (さらに低めの閾値)
    const ULTIMATE_IMAX_COMPRESSOR_RATIO_ADVANCED = 15; // :1 (極めて高めの圧縮比)
    const ULTIMATE_IMAX_COMPRESSOR_ATTACK_ADVANCED = 0.0001; // seconds (瞬間的なアタック)
    const ULTIMATE_IMAX_COMPRESSOR_RELEASE_ADVANCED = 1.2; // seconds (非常に長めのリリース)
    const ULTIMATE_IMAX_COMPRESSOR_KNEE_ADVANCED = 25; // dB (極めて広めのニー)

     const ULTIMATE_IMAX_EXPANDER_BOOST_GAIN_ADVANCED = 5.0; // 簡易エキスパンダー風ゲインを最大強調

     // ピークリミッター設定 - さらにアグレッシブに
     const PEAK_LIMITER_THRESHOLD_DB_ADVANCED = 0; // dB
     const PEAK_LIMITER_ATTACK_ADVANCED = 0; // seconds
     const PEAK_LIMITER_RELEASE_ADVANCED = 0.005; // seconds (極めて速いリリース)


    // 簡易ステレオワイド設定例 - v8.0ベース維持
    const STEREO_WIDE_DELAY_MS = 30;
    const STEREO_WIDE_LEFT_GAIN = 1.0;
    const STEREO_WIDE_RIGHT_GAIN = 0.85;

    // 簡易リバーブ設定例 (Convolutionと併用するため、補助的に調整) - v8.0ベース維持
    const REVERB_DELAY_SEC = 0.05;
    const REVERB_FEEDBACK = 0.4;
    const REVERB_WET_GAIN = 0.1;


    // dB値をゲイン値に変換するヘルパー関数
    function dBToGain(db) {
        return Math.pow(10, db / 20);
    }

     // little-endian float PCM ArrayBuffer を Float32Array にデコードする簡易関数例 (実際の処理フローでは使用されません)
     function decodeLittleEndianFloat32(arrayBuffer) {
         console.log(`Attempting to decode ArrayBuffer of length ${arrayBuffer.byteLength} as little-endian Float32 PCM.`);
         if (arrayBuffer.byteLength % 4 !== 0) {
             console.error("ArrayBuffer byte length is not a multiple of 4. Cannot decode as Float32.");
             return new Float32Array(0);
         }
         const dataView = new DataView(arrayBuffer);
         const numSamples = arrayBuffer.byteLength / 4;
         const float32Array = new Float32Array(numSamples);

         for (let i = 0; i < numSamples; i++) {
             // DataView.getFloat32(byteOffset, littleEndian)
             float32Array[i] = dataView.getFloat32(i * 4, true); // true for little-endian
         }
         console.log(`Successfully decoded ${numSamples} Float32 samples.`);
         return float32Array;
     }


    // --- 音響物理学応用モデル計算関数 ---
    // 評価計算部分は維持 (ログ出力なし)

    /**
     * Cross-sectional area of horn at distance x
     * S(x) = S0 * exp(m * x)
     */
    function S(x) {
        return HORN_S0 * Math.exp(HORN_M * x);
    }

    /**
     * Acoustic impedance at distance x
     * Z(x) = (rho0 * c) / S(x)
     */
    function Z(x) {
        return (RHO0 * C) / S(x);
    }

    /**
     * Exact plane wave pressure at (x, t)
     * p(x, t) = P0 * exp(j*(omegat - kx))
     * Returns a complex number as { re, im }
     */
    function pPlane(x, t, frequency, sourceAmplitude_P0) {
        const omega = 2 * Math.PI * frequency;
        const k = omega / C;
        const phase = omega * t - k * x;
        return {
            re: sourceAmplitude_P0 * Math.cos(phase),
            im: sourceAmplitude_P0 * Math.sin(phase)
        };
    }

    /**
     * Approximate pressure including area-expansion attenuation
     * p_approx(x, t) = P0 * sqrt(S0/S(x)) * exp(j*(omegat - kx))
     */
    function pApprox(x, t, frequency, sourceAmplitude_P0) {
        const factor = Math.sqrt(HORN_S0 / S(x));
        const omega = 2 * Math.PI * frequency;
        const k = omega / C;
        const phase = omega * t - k * x;
        return {
            re: sourceAmplitude_P0 * factor * Math.cos(phase),
            im: sourceAmplitude_P0 * factor * Math.sin(phase)
        };
    }


    // 音響管の共鳴特性計算関数 (簡易モデル)
    function calculateTubeResonanceGain(frequency, resonanceFrequency, speedOfSound, Q = 10) {
        const delta_f = Math.abs(frequency - resonanceFrequency);
        const bandwidth = resonanceFrequency / Q;
        return 1.0 / (1.0 + Math.pow(delta_f / (bandwidth / 2), 2));
    }

     // サブハーモニック合成ノードを作成する関数
    function createSubharmonicNode(audioCtx, ratio, gainValue) {
        const osc = audioCtx.createOscillator();
        osc.type = 'sine';
        const gain = audioCtx.createGain();
        gain.gain.value = gainValue;
        osc.connect(gain);
        return { osc, gain };
    }

     // 基本周波数を推定する関数 (自己相関法)
     function estimateFundamental(analyser) {
        if (!analyser) {
             console.warn("AnalyserNode not provided for fundamental frequency estimation.");
             return 0; // 推定失敗
        }
        const data = new Float32Array(analyser.fftSize);
        analyser.getFloatTimeDomainData(data);

        let bestOffset = -1;
        let bestCorr = -Infinity;
        const sampleRate = analyser.context.sampleRate;
        const maxOffset = Math.min(Math.floor(sampleRate / 20), data.length / 2); // 最低20Hzまで
        const minOffset = Math.max(2, Math.floor(sampleRate / 500)); // 最高500Hzまで

        for (let off = minOffset; off < maxOffset; off++) {
            let corr = 0;
            for (let i = 0; i < data.length - off; i++) {
                corr += data[i] * data[i + off];
            }
            if (corr > bestCorr) { bestCorr = corr; bestOffset = off; }
        }

        if (bestCorr > 0.1 * data.length && bestOffset > 0) { // 適当な閾値とオフセットチェック
            const fundamental = sampleRate / bestOffset;
             console.log(`Estimated fundamental frequency: ${fundamental.toFixed(2)} Hz (Offset: ${bestOffset})`);
            return fundamental;
        } else {
             console.log("Fundamental frequency estimation failed (correlation too low or offset zero).");
             return 0; // 推定失敗
        }
    }

    // 簡易合成IRデータを生成する関数
     function createSyntheticIR(audioContext, duration, decayRate, numReflections, maxDelay) {
         const sampleRate = audioContext.sampleRate;
         const numSamples = Math.floor(duration * sampleRate);
         const buffer = audioContext.createBuffer(2, numSamples, sampleRate); // ステレオIR

         for (let channel = 0; channel < buffer.numberOfChannels; channel++) {
             const nowBuffering = buffer.getChannelData(channel);
             nowBuffering.fill(0); // バッファをゼロで初期化

             // 直接音 (最初のピーク)
             if (numSamples > 0) {
                  nowBuffering[0] = 1.0; // 最初のサンプルを1.0に

                  // 簡易的なランダム反射音を加える
                  for(let i = 0; i < numReflections; i++) {
                       const delaySamples = Math.floor(Math.random() * maxDelay * sampleRate / 1000); // ランダムなディレイ（サンプル数）
                       const gain = (Math.random() * 0.4 + 0.1) * Math.exp(-delaySamples / sampleRate * decayRate); // ランダムなゲインと減衰
                       if (delaySamples < numSamples) {
                           nowBuffering[delaySamples] += gain * (channel === 0 ? 1 : (Math.random() * 0.5 + 0.5)); // 左右チャンネルで少し異なるゲイン
                       }
                  }
             }
         }
         return buffer;
     }
     // Function to create channel independent processing nodes (trim, delay, simplified phase emulation)
     // Input: audioContext, inputNode, channelCount
     // Output: outputNode after channel independent processing
    function createChannelIndependentProcessing(audioContext, inputNode, channelCount) {
        console.log(`Creating channel independent processing nodes for ${channelCount} channels.`);
        const channelSplitter = audioContext.createChannelSplitter(channelCount);
        const channelMerger = audioContext.createChannelMerger(channelCount);

        inputNode.connect(channelSplitter);

        for (let i = 0; i < channelCount; i++) {
            // トリム調整
            const trimGain = audioContext.createGain();
            // 設定値のチャンネル数と実際のチャンネル数を比較し、足りない場合はデフォルト値を使用
            const trimDb = (i < CHANNEL_TRIM_DB.length) ? CHANNEL_TRIM_DB[i] : 0;
            trimGain.gain.setValueAtTime(dBToGain(trimDb), audioContext.currentTime);
            channelSplitter.connect(trimGain, i); // スプリッターのi番目出力をゲイン入力へ

            // 遅延補正
            const delayNode = audioContext.createDelay(1.0); // 最大遅延1秒
            const delayMs = (i < CHANNEL_DELAY_MS.length) ? CHANNEL_DELAY_MS[i] : 0;
            delayNode.delayTime.setValueAtTime(delayMs / 1000, audioContext.currentTime);
            trimGain.connect(delayNode); // ゲイン出力を遅延入力へ

            // 位相整合（簡易模倣）のためのAllpass Filter
            // Allpass Filterは周波数応答をフラットに保ちつつ位相を操作する
            const allpassFilter = audioContext.createBiquadFilter();
            allpassFilter.type = 'allpass';
            const allpassFreq = (i < CHANNEL_PHASE_ALLPASS_FREQ.length) ? CHANNEL_PHASE_ALLPASS_FREQ[i] : 500; // デフォルト500Hz
            const allpassQ = (i < CHANNEL_PHASE_ALLPASS_Q.length) ? CHANNEL_PHASE_ALLPASS_Q[i] : 1.0; // デフォルト1.0
            allpassFilter.frequency.setValueAtTime(allpassFreq, audioContext.currentTime);
            allpassFilter.Q.setValueAtTime(allpassQ, audioContext.currentTime);
            delayNode.connect(allpassFilter); // 遅延出力をAllpass入力へ

            // 各チャンネル処理の出力をマージャーへ
            allpassFilter.connect(channelMerger, 0, i); // Allpass出力をマージャーのi番目入力へ
        }

        console.log("Channel independent processing nodes created.");
        return channelMerger; // マージャーの出力がこの関数の出力ノード
    }
    // Function to create advanced spatial and dynamics processing nodes (IMAX simulation, etc.)
    // This includes: Stereo Wide, M/S, EQ, Expander, Physics Filters, Spatial Mix (Simple Reflections, Convolution, Simple Reverb), Compressor, Subharmonic, Peak Limiter, Final Gain
    // It attempts to conceptually simulate: Spatiotemporal energy optimization, Vector sum synthesis, Wavefront correction (partially via channel control), Surround focus layer
    // Input: audioContext, inputNode, channelCount, fundamentalFrequency
    // Output: outputNode before connecting to destination (typically finalGainNode)
    function createAdvancedSpatialDynamicsProcessing(audioContext, inputNode, channelCount, fundamentalFrequency) {
        console.log("Creating advanced spatial and dynamics processing nodes (IMAX Simulation).");

        let currentNode = inputNode; // Start from the input node of this function

        // 簡易ステレオワイド処理はステレオ(2ch)を前提とする
        if (channelCount >= 2) {
            const splitterSimStereo_rt = audioContext.createChannelSplitter(2); // ステレオ分解
            const mergerSimStereo_rt = audioContext.createChannelMerger(2); // ステレオ合成

            currentNode.connect(splitterSimStereo_rt);

            const delayRightSimStereo_rt = audioContext.createDelay(1.0);
            delayRightSimStereo_rt.delayTime.setValueAtTime(STEREO_WIDE_DELAY_MS / 1000, audioContext.currentTime);
            const gainLeftSimStereo_rt = audioContext.createGain();
            gainLeftSimStereo_rt.gain.setValueAtTime(STEREO_WIDE_LEFT_GAIN, audioContext.currentTime);
            const gainRightSimStereo_rt = audioContext.createGain();
            gainRightSimStereo_rt.gain.setValueAtTime(STEREO_WIDE_RIGHT_GAIN, audioContext.currentTime);

            if (splitterSimStereo_rt.numberOfOutputs >= 2) {
                 splitterSimStereo_rt.connect(gainLeftSimStereo_rt, 0);
                 gainLeftSimStereo_rt.connect(mergerSimStereo_rt, 0, 0);
                 splitterSimStereo_rt.connect(delayRightSimStereo_rt, 1);
                 delayRightSimStereo_rt.connect(gainRightSimStereo_rt);
                 gainRightSimStereo_rt.connect(mergerSimStereo_rt, 0, 1);
            }
            currentNode = mergerSimStereo_rt; // ステレオワイドの出力が次へ
        } else {
            console.warn("Source has less than 2 channels. Skipping stereo wide processing. Using StereoPanner for mono->stereo conversion.");
            // モノラルソースの場合、簡易的にモノラルをステレオに複製して次へ渡す
            const stereoPanner = audioContext.createStereoPanner();
            stereoPanner.pan.setValueAtTime(0, audioContext.currentTime); // センターに定位
            currentNode.connect(stereoPanner);
            currentNode = stereoPanner; // ステレオパンナーの出力が次へ (2ch)
        }


        // 前段の処理出力 -> Mid/Side分解 -> ...
        // M/S処理はステレオ(2ch)を前提とする
        // 前段の処理で出力が2chになっていることを期待
        const splitterMS_rt = audioContext.createChannelSplitter(2); // ステレオ分解
        const mergerMS_rt = audioContext.createChannelMerger(2); // ステレオ合成

        currentNode.connect(splitterMS_rt);

        // Mid (L+R)/2, Side (L-R)/2 を計算するゲインノード
        const midGain_calc = audioContext.createGain(); midGain_calc.gain.setValueAtTime(0.5, audioContext.currentTime);
        const sideGain_calc = audioContext.createGain(); sideGain_calc.gain.setValueAtTime(0.5, audioContext.currentTime);
        const negSideGain_calc = audioContext.createGain(); negSideGain_calc.gain.setValueAtTime(-0.5, audioContext.currentTime);

        splitterMS_rt.connect(midGain_calc, 0); // L -> Mid
        splitterMS_rt.connect(midGain_calc, 1); // R -> Mid
        splitterMS_rt.connect(sideGain_calc, 0); // L -> Side
        splitterMS_rt.connect(negSideGain_calc, 1); // R -> -Side
        negSideGain_calc.connect(sideGain_calc); // -Side -> Side (L-R)

        // Mid/Side成分に対するゲイン調整
        // ベクトル方向の合成（Vector Sum Synthesis）の概念をMid/Side処理の強調で模倣
        const midProcessGain_rt = audioContext.createGain();
        midProcessGain_rt.gain.setValueAtTime(dBToGain(MID_GAIN_DB), audioContext.currentTime);
        const sideProcessGain_rt = audioContext.createGain();
        sideProcessGain_rt.gain.setValueAtTime(dBToGain(SIDE_GAIN_DB), audioContext.currentTime);

        midGain_calc.connect(midProcessGain_rt);
        sideGain_calc.connect(sideProcessGain_rt);

        // Mid/Side成分を左右チャンネルに再合成
        const gainOne_rt = audioContext.createGain(); gainOne_rt.gain.setValueAtTime(1.0, audioContext.currentTime);
        const gainNegOne_rt = audioContext.createGain(); gainNegOne_rt.gain.setValueAtTime(-1.0, audioContext.currentTime);

        // Left = Mid + Side
        midProcessGain_rt.connect(mergerMS_rt, 0, 0);
        sideProcessGain_rt.connect(mergerMS_rt, 0, 0);

        // Right = Mid - Side
        midProcessGain_rt.connect(mergerMS_rt, 0, 1);
        sideProcessGain_rt.connect(gainNegOne_rt, 0); // Side成分に-1をかけてから
        gainNegOne_rt.connect(mergerMS_rt, 0, 1); // Rightチャンネルに接続

        currentNode = mergerMS_rt; // M/S処理後の出力が次へ


        // 前段の処理出力 -> EQ補正群 -> 簡易エキスパンダー風Gain -> 物理応用フィルター群 -> ...

        // EQ補正群 (IMAX向け調整)
        // 時空間的なエネルギー分布の最適化を図る (主にEQ/Comp/Spatial処理で実現を試みる)
        const imaxEQs_rt = ULTIMATE_IMAX_EQ_SETTINGS_ADVANCED.map(settings => {
             const eq = audioContext.createBiquadFilter();
             eq.type = settings.type;
             eq.frequency.setValueAtTime(settings.freq, audioContext.currentTime);
             eq.gain.setValueAtTime(settings.gain, audioContext.currentTime);
             if (settings.Q) eq.Q.setValueAtTime(settings.Q, audioContext.currentTime);
             return eq;
         });

        imaxEQs_rt.forEach(eq => {
             currentNode.connect(eq);
             currentNode = eq;
         });

        // 簡易エキスパンダー風ゲインノード
        const expanderGain_rt = audioContext.createGain();
        expanderGain_rt.gain.setValueAtTime(ULTIMATE_IMAX_EXPANDER_BOOST_GAIN_ADVANCED, audioContext.currentTime);
        currentNode.connect(expanderGain_rt);
        currentNode = expanderGain_rt;


        // 物理応用フィルター群を接続
        const hornCutoffFilter_rt = audioContext.createBiquadFilter();
        hornCutoffFilter_rt.type = 'highpass';
        hornCutoffFilter_rt.frequency.setValueAtTime(HORN_CUTOFF_FREQ_CALCULATED, audioContext.currentTime);
        hornCutoffFilter_rt.Q.setValueAtTime(0.5, audioContext.currentTime);

        const tubeResonanceCompFilter_rt = audioContext.createBiquadFilter();
        tubeResonanceCompFilter_rt.type = 'notch';
        tubeResonanceCompFilter_rt.frequency.setValueAtTime(TUBE_RESONANCE_FREQ, audioContext.currentTime);
        tubeResonanceCompFilter_rt.Q.setValueAtTime(10.0, audioContext.currentTime);

        currentNode.connect(hornCutoffFilter_rt);
        hornCutoffFilter_rt.connect(tubeResonanceCompFilter_rt);
        currentNode = tubeResonanceCompFilter_rt;


        // 簡易空間処理群 (簡易反射, Convolution, 簡易リバーブ) とドライ信号のミックス
        // 波面整合処理（Wavefront Correction）の概念をチャンネル独立遅延/位相模倣で一部模倣
        // サラウンド・フォーカス・レイヤ（Surround Focus Layer）の概念をM/S強調と空間処理で一部模倣

        // 物理応用フィルター出力から分岐して各空間処理へ
        let spatialProcessInput = currentNode;

        // 各成分を最終ミックスMergerに接続
        // Mergerの出力チャンネル数は、最終的にDestinationに接続されるチャンネル数（通常は2、つまりステレオ）に合わせる
        const finalMixMerger = audioContext.createChannelMerger(2);


        // ドライ信号（Convolutionのドライゲイン）を接続
        // spatialProcessInput (物理応用フィルター出力) -> dryGain_rt -> finalMixMerger
        const dryGain_rt = audioContext.createGain();
        dryGain_rt.gain.setValueAtTime(CONVOLVER_DRY_GAIN, audioContext.currentTime);
        spatialProcessInput.connect(dryGain_rt);
        // ドライ信号をステレオとしてMergerに接続 (ソースのチャンネル数に関わらず2ch出力に変換)
        dryGain_rt.connect(finalMixMerger, 0, 0);
        dryGain_rt.connect(finalMixMerger, 0, 1);


        // -> 簡易反射ノード群 (ミックス後、最終ミックスMergerへ)
        // spatialProcessInput -> Delay -> Gain -> earlyReflectionMerger_rt -> finalMixMerger
        const earlyReflectionDelays_rt = EARLY_REFLECTION_DELAYS_MS.map(delayMs => {
             const delay = audioContext.createDelay(1.0);
             delay.delayTime.setValueAtTime(delayMs / 1000, audioContext.currentTime);
             return delay;
         });
        const earlyReflectionGains_rt = EARLY_REFLECTION_GAINS.map(gainValue => {
             const gain = audioContext.createGain();
             gain.gain.setValueAtTime(gainValue, audioContext.currentTime);
             return gain;
         });
         // 簡易反射の出力はステレオとしてMergerに結合することを想定
         const earlyReflectionMerger_rt = audioContext.createChannelMerger(2);

        earlyReflectionDelays_rt.forEach((delayNode, index) => {
             if (index < earlyReflectionGains_rt.length) {
                 spatialProcessInput.connect(delayNode);
                 delayNode.connect(earlyReflectionGains_rt[index]);
                 // 簡易反射の出力はステレオとして結合
                 earlyReflectionGains_rt[index].connect(earlyReflectionMerger_rt, 0, 0);
                 earlyReflectionGains_rt[index].connect(earlyReflectionMerger_rt, 0, 1);
             }
         });
        // 簡易反射のミックス出力を最終ミックスMergerへ
        earlyReflectionMerger_rt.connect(finalMixMerger, 0, 0);
        earlyReflectionMerger_rt.connect(finalMixMerger, 0, 1);


        // -> Convolution Node (ウェット信号)
        // spatialProcessInput -> convolverNode -> convolverWetGain -> finalMixMerger
        const convolverNode = audioContext.createConvolver();
         try {
             convolverNode.buffer = createSyntheticIR(
                 audioContext,
                 CONVOLVER_IR_DURATION_SEC,
                 CONVOLVER_IR_DECAY_RATE,
                 CONVOLVER_IR_NUM_REFLECTIONS,
                 EARLY_REFLECTION_DELAYS_MS[EARLY_REFLECTION_DELAYS_MS.length - 1] || 50
             );
              console.log("Synthetic Impulse Response created and loaded into ConvolverNode.");
         } catch (e) {
             console.error("Failed to create or load synthetic IR for ConvolverNode:", e);
         }

        const convolverWetGain = audioContext.createGain();
        convolverWetGain.gain.setValueAtTime(CONVOLVER_WET_GAIN, audioContext.currentTime);

        if (convolverNode.buffer) { // IRがロードできていれば接続
            spatialProcessInput.connect(convolverNode);
            convolverNode.connect(convolverWetGain);
            // Convolutionのウェット信号をステレオとしてMergerに接続
            convolverWetGain.connect(finalMixMerger, 0, 0);
            convolverWetGain.connect(finalMixMerger, 0, 1);
        } else {
            console.warn("ConvolutionNode buffer is not set. Skipping Convolution.");
        }


        // -> 簡易リバーブ (ウェット信号)
        // spatialProcessInput -> reverbDelay_rt -> reverbFeedbackGain_rt -> reverbDelay_rt (Feedback Loop)
        // spatialProcessInput -> reverbDelay_rt -> reverbWetGain_rt -> finalMixMerger
        const reverbDelay_rt = audioContext.createDelay(1.0);
        reverbDelay_rt.delayTime.setValueAtTime(REVERB_DELAY_SEC, audioContext.currentTime);
        const reverbFeedbackGain_rt = audioContext.createGain();
        reverbFeedbackGain_rt.gain.setValueAtTime(REVERB_FEEDBACK, audioContext.currentTime);
        const reverbWetGain_rt = audioContext.createGain();
        reverbWetGain_rt.gain.setValueAtTime(REVERB_WET_GAIN, audioContext.currentTime);

        spatialProcessInput.connect(reverbDelay_rt); // リバーブ入力
        reverbDelay_rt.connect(reverbFeedbackGain_rt); // フィードバックループ
        reverbFeedbackGain_rt.connect(reverbDelay_rt); // フィードバックループ完了

        reverbDelay_rt.connect(reverbWetGain_rt); // ウェット信号取得
        // 簡易リバーブウェット信号をステレオとしてMergerに接続
        reverbWetGain_rt.connect(finalMixMerger, 0, 0);
        reverbWetGain_rt.connect(finalMixMerger, 0, 1);


        // 全ての空間処理成分が finalMixMerger に接続された
        currentNode = finalMixMerger; // finalMixMerger の出力が次へ


        // ミックス出力 -> 全体のダイナミクス調整 -> ピークリミッター -> ラウドネス調整Gain
        // 時空間的なエネルギー分布の最適化を図る (主にEQ/Comp/Spatial処理で実現を試みる)
        // サラウンド・フォーカス・レイヤ（Surround Focus Layer）の概念をM/S強調と空間処理、全体ダイナミクスで一部模倣

        // 全体のダイナミクス調整用コンプレッサー
        const compressor_rt = audioContext.createDynamicsCompressor();
        compressor_rt.threshold.setValueAtTime(ULTIMATE_IMAX_COMPRESSOR_THRESHOLD_ADVANCED, audioContext.currentTime);
        compressor_rt.ratio.setValueAtTime(ULTIMATE_IMAX_COMPRESSOR_RATIO_ADVANCED, audioContext.currentTime);
        compressor_rt.attack.setValueAtTime(ULTIMATE_IMAX_COMPRESSOR_ATTACK_ADVANCED, audioContext.currentTime);
        compressor_rt.release.setValueAtTime(ULTIMATE_IMAX_COMPRESSOR_RELEASE_ADVANCED, audioContext.currentTime);
        compressor_rt.knee.setValueAtTime(ULTIMATE_IMAX_COMPRESSOR_KNEE_ADVANCED, audioContext.currentTime);
        currentNode.connect(compressor_rt);
        currentNode = compressor_rt;


        // サブハーモニック信号をコンプレッサーの入力に加算
        let subOsc_rt = null; // オシレーターをこのスコープで宣言
        let subGain_rt = null; // ゲインもここで宣言
         if (fundamentalFrequency > 0) {
             const subNodes = createSubharmonicNode(audioContext, SUBHARMONIC_RATIO, SUBHARMONIC_GAIN);
             subOsc_rt = subNodes.osc; // 参照を保持
             subGain_rt = subNodes.gain; // 参照を保持
             subOsc_rt.frequency.setValueAtTime(fundamentalFrequency / SUBHARMONIC_RATIO, audioContext.currentTime);
             // サブハーモニック信号をコンプレッサーの入力に接続
             subGain_rt.connect(compressor_rt);
              console.log("Subharmonic signal connected to compressor input.");
              // Contextがrunningになったらオシレーターを開始
              if (audioContext.state === 'running') {
                  subOsc_rt.start();
              } else {
                  // Contextがsuspendedの場合は resume 時に開始
                  audioContext.resume().then(() => {
                       if (subOsc_rt && subOsc_rt.state === 'pending') {
                          subOsc_rt.start();
                           console.log("Subharmonic oscillator started after context resumed.");
                       }
                  }).catch(e => console.error("Failed to resume context for subharmonic start:", e));
              }
         } else {
              console.warn("Fundamental frequency not estimated or is zero. Subharmonic oscillator will not be used.");
         }


        // ピークリミッター
        const peakLimiter_rt = audioContext.createDynamicsCompressor();
        peakLimiter_rt.threshold.setValueAtTime(PEAK_LIMITER_THRESHOLD_DB_ADVANCED, audioContext.currentTime);
        peakLimiter_rt.ratio.setValueAtTime(100, audioContext.currentTime);
        peakLimiter_rt.attack.setValueAtTime(PEAK_LIMITER_ATTACK_ADVANCED, audioContext.currentTime);
        peakLimiter_rt.release.setValueAtTime(PEAK_LIMITER_RELEASE_ADVANCED, audioContext.currentTime);
        peakLimiter_rt.knee.setValueAtTime(0, audioContext.currentTime);
        currentNode.connect(peakLimiter_rt);
        currentNode = peakLimiter_rt;


        // ラウドネス調整用ゲイン
        const finalGainNode_rt = audioContext.createGain();
        finalGainNode_rt.gain.setValueAtTime(1.0, audioContext.currentTime);
        currentNode.connect(finalGainNode_rt);
        currentNode = finalGainNode_rt;

        console.log("Advanced spatial and dynamics processing nodes created.");
        return currentNode; // 最終ゲインノードがこの関数の出力ノード
    }

    // --- AudioWorkletProcessor definition (Advanced Concepts & Dynamic Interpolation) ---
    // LocalEnhance処理を高度化・リアルタイム向けに再設計（動的補間、概念応用、小数点処理）
    const processorCodeOptimized = `
    class LocalEnhanceProcessor extends AudioWorkletProcessor {
        // Worklet内の設定値
        LE_CHUNK_SIZE = ${LE_CHUNK_SIZE_WORKLET};
        LE_RMS_THRESHOLD = ${LE_RMS_THRESHOLD_WORKLET};
        EW_WAVELENGTH_FACTOR = ${EW_WAVELENGTH_FACTOR};
        EW_ENHANCEMENT_FACTOR = ${EW_ENHANCEMENT_FACTOR};
        GAMMA_L = ${GAMMA_L};
        GAMMA_T = ${GAMMA_T};
        GAMMA_MU0 = ${GAMMA_MU0};
        GAMMA_KAPPA = ${GAMMA_KAPPA};
        GAMMA_ALPHAP = ${GAMMA_ALPHAP};
        GAMMA_THRESHOLD = ${GAMMA_THRESHOLD_WORKLET};
        GAMMA_INTERPOLATION_FACTOR = ${GAMMA_INTERPOLATION_FACTOR};
        GAMMA_DYNAMIC_GAIN_MULTIPLIER = ${GAMMA_DYNAMIC_GAIN_MULTIPLIER};

        // グローバー反復回数概念応用設定
        GROVER_ITERATION_BASE = ${GROVER_ITERATION_BASE};
        GROVER_OPTIMIZATION_FACTOR = ${GROVER_OPTIMIZATION_FACTOR};

        // 小数点処理設定
        FIXED_PRECISION_DECIMALS = ${FIXED_PRECISION_DECIMALS};
        precisionFactor = Math.pow(10, ${FIXED_PRECISION_DECIMALS});


        // チャンク処理と補間のためのバッファ（過去のサンプルを保持）
        // Float32 PCM バッファを明示的に管理・応用
        chunkBuffer = []; // チャンネルごとの Float32Array バッファ
        bufferLength = 128; // 標準の処理バッファ長 (Workletの入力/出力サイズ)
        // RMS計算や補間に必要な過去のサンプル数 (チャンクサイズの2倍 + 補間に必要な数)
        historyBufferLength = this.LE_CHUNK_SIZE * 2 + 4; // 4は3次補間に必要な前後サンプルの余裕

        constructor() {
            super();
            this.chunkBuffer = new Array(this.numberOfInputs[0]);
            for(let i = 0; i < this.numberOfInputs[0]; i++) {
                 this.chunkBuffer[i] = new Float32Array(this.bufferLength + this.historyBufferLength);
                 this.chunkBuffer[i].fill(0); // バッファをゼロで初期化
            }
             // console.log('[Worklet] Processor initialized.');
        }

        // 小数点以下第N位に丸める関数 (toFixed(2) 模倣)
        roundToPrecision(value) {
            return Math.round(value * this.precisionFactor) / this.precisionFactor;
        }

        // 簡易RMS計算
        calculateRMS(buffer) {
            let sum = 0;
            for (let i = 0; i < buffer.length; i++) {
                sum += buffer[i] * buffer[i];
            }
            return Math.sqrt(sum / buffer.length);
        }

        // 簡易3次補間
        cubicInterpolate(y0,y1,y2,y3,mu) {
            const a0=y3-y2-y0+y1;
            const a1=y0-y1-a0;
            const a2=y2-y0;
            const a3=y1;
            return a0*mu*mu*mu + a1*mu*mu + a2*mu + a3;
        }

        // 信号の状態判定（簡易的な立ち上がり検出）
        // 前のサンプルとの差分を見て急峻さを判定
        isSignalAttack(buffer, currentSampleIndexInBuffer, threshold = 0.01) {
            if (currentSampleIndexInBuffer === 0) return false;
            const currentSample = buffer[currentSampleIndexInBuffer];
            const prevSample = buffer[currentSampleIndexInBuffer - 1];
            return Math.abs(currentSample - prevSample) > threshold;
        }

        // グローバー反復回数概念応用による動的パラメータ計算（概念的な応用）
        // 信号のエネルギーや特徴（立ち上がりなど）に応じて、概念的な「反復回数」を調整し、
        // それを元に動的な補間強度やゲイン乗数を計算
        calculateDynamicOptimizationFactor(rms, isAttack) {
            // RMSが低いほど、または立ち上がりがあるほど、強調や最適化の必要性が高いと解釈
            let optimizationBase = this.GROVER_ITERATION_BASE;

            if (rms < this.LE_RMS_THRESHOLD * 2) { // RMSが低い場合
                optimizationBase += (this.LE_RMS_THRESHOLD * 2 - rms) * 100; // RMSが低いほどベースを増加
            }

            if (isAttack) { // 立ち上がりの場合
                optimizationBase += 5; // 立ち上がり点でさらに強調
            }

            // 概念的な反復回数を基に、動的な強調ファクターを計算
            // 例: sqrt(optimizationBase) のような非線形関数を使用
            // これを0-1の範囲に正規化し、最適化の度合い (GROVER_OPTIMIZATION_FACTOR) を適用
            const dynamicFactor = Math.sqrt(optimizationBase) * this.GROVER_OPTIMIZATION_FACTOR;

            // ファクターを適切な範囲に制限
            return Math.max(0, Math.min(1.0, dynamicFactor));
        }


        process(inputs, outputs) {
            const input = inputs[0];
            const output = outputs[0];
            const channelCount = input ? input.length : 0;

            if (channelCount > 0) {
                for(let ch = 0; ch < channelCount; ch++) {
                    const inputChannel = input[ch];
                    const outputChannel = output[ch];

                    if (inputChannel && outputChannel) {
                        const inputLength = inputChannel.length; // 通常128

                        // Float32 PCM バッファの更新（新しいデータをバッファに追加し、古いデータを捨てる）
                        const buffer = this.chunkBuffer[ch];
                        // 古いデータをhistoryBufferLengthの開始位置に詰める
                        buffer.set(buffer.subarray(inputLength), 0);
                        // 新しい入力データをhistoryBufferLength以降に追加
                        buffer.set(inputChannel, this.historyBufferLength);


                        // Workletの処理ブロックごとにLocalEnhance処理を適用
                        for (let i = 0; i < inputLength; i++) {
                            // バッファ全体における現在の入力サンプルの位置
                            const currentSampleIndexInBuffer = this.historyBufferLength + i;

                            // ローカルエンハンスメント処理（概念応用、動的補間、小数点処理）

                            // RMS計算用のチャンク
                             const rmsChunkStartIndex = Math.max(0, currentSampleIndexInBuffer - Math.floor(this.LE_CHUNK_SIZE / 2));
                             const rmsChunkEndIndex = Math.min(buffer.length, rmsChunkStartIndex + this.LE_CHUNK_SIZE);
                             const rmsChunk = buffer.subarray(rmsChunkStartIndex, rmsChunkEndIndex);
                            const rms = this.calculateRMS(rmsChunk);

                            let processedSample = inputChannel[i]; // デフォルトは元のサンプル

                            // RMS条件を満たすかどうかで分岐
                            if (rms < this.LE_RMS_THRESHOLD) {
                                // RMS条件を満たす場合

                                // 信号状態判定（立ち上がり検出）
                                 const isAttack = this.isSignalAttack(buffer, currentSampleIndexInBuffer);

                                // グローバー反復概念応用による動的パラメータ計算
                                 const dynamicFactor = this.calculateDynamicOptimizationFactor(rms, isAttack);

                                // 弾性波動模倣を適用
                                const preIndex_ew = Math.max(0, currentSampleIndexInBuffer - 1);
                                const targetIndex_ew = currentSampleIndexInBuffer;
                                const preValue_ew = buffer[preIndex_ew];
                                const targetValue_ew = buffer[targetIndex_ew];
                                 // Workletバッファ内の相対位置 i を使用 (0-127) を0-1に正規化してw計算
                                const w_ew = (i / inputLength) * this.EW_ENHANCEMENT_FACTOR;
                                let sampleAfterEW = preValue_ew * (1 - w_ew) + targetValue_ew * w_ew;


                                // Gamma条件を満たすかどうかをチェック
                                 const mu_i = Math.abs(sampleAfterEW); // 弾性波動適用後のサンプル値でGamma条件をチェック
                                 let gamma = 0;
                                 if (mu_i > 1e-9) {
                                     const logArg = Math.abs((Math.sqrt(this.GAMMA_MU0) - Math.sqrt(mu_i)) / (Math.sqrt(this.GAMMA_MU0) + Math.sqrt(mu_i)));
                                     if (logArg > 1e-9) {
                                         gamma = -1 / this.GAMMA_L * Math.sqrt(this.GAMMA_T / mu_i) * Math.log(logArg) * (1 + this.GAMMA_KAPPA * this.GAMMA_ALPHAP / (this.GAMMA_L * this.GAMMA_L));
                                     }
                                 }
                                 const g = gamma;

                                if (Math.abs(g) > this.GAMMA_THRESHOLD) {
                                     // Gamma条件を満たす場合、動的補間による波形調整を適用
                                     // 現在のサンプルとその前後のサンプルを使用して補間値を計算
                                     // 簡易的な4点補間を応用
                                     const y0_idx = Math.max(0, currentSampleIndexInBuffer - 2);
                                     const y1_idx = Math.max(0, currentSampleIndexInBuffer - 1);
                                     const y2_idx = Math.min(buffer.length - 1, currentSampleIndexInBuffer + 1);
                                     const y3_idx = Math.min(buffer.length - 1, currentSampleIndexInBuffer + 2);

                                     const y0 = buffer[y0_idx];
                                     const y1 = buffer[y1_idx];
                                     const y2 = buffer[y2_idx];
                                     const y3 = buffer[y3_idx];

                                     // 補間ポイントの調整 (例: 現在のサンプルと次のサンプルの中間mu=0.5)
                                     const mu = 0.5; // 静的な中間点補間
                                     // 将来的には、動的ファクターに基づきmuを動的に調整することも考えられる
                                     // const mu = 0.5 + (dynamicFactor * 0.4); // 簡易的な動的mu調整例

                                     const interpolatedValue = this.cubicInterpolate(y0, y1, targetValue_ew, y2, mu); // 弾性波動適用前の元の値を補間に使用

                                     // 弾性波動適用後のサンプルと補間値を、動的ファクターと設定に基づいてブレンド
                                     // 動的ファクターが高いほど補間値の影響を強くする
                                     const interpolationFactor = this.GAMMA_INTERPOLATION_FACTOR * dynamicFactor; // 動的にブレンド比率を調整
                                     let sampleAfterInterpolation = sampleAfterEW * (1.0 - interpolationFactor) + interpolatedValue * interpolationFactor;

                                     // 動的ゲイン調整（グローバー概念応用）
                                     // 動的ファクターが高いほどゲインを増加させる
                                     const dynamicGain = 1.0 + (this.GAMMA_DYNAMIC_GAIN_MULTIPLIER - 1.0) * dynamicFactor;
                                     processedSample = sampleAfterInterpolation * dynamicGain;


                                     // 特定の計算結果に対して小数点第2位に丸める (模倣)
                                     // 例として、最終的な出力サンプル値を丸める
                                     processedSample = this.roundToPrecision(processedSample);

                                } else {
                                     // Gamma条件を満たさない場合は、弾性波動適用後のサンプルをそのまま使用
                                     processedSample = sampleAfterEW;
                                     // 特定の計算結果に対して小数点第2位に丸める (模倣)
                                     // 例として、最終的な出力サンプル値を丸める
                                     processedSample = this.roundToPrecision(processedSample);
                                }

                            } else {
                                // RMS条件を満たさない場合は、元の入力サンプルをそのまま使用
                                processedSample = inputChannel[i];
                                // 特定の計算結果に対して小数点第2位に丸める (模倣)
                                // 例として、最終的な出力サンプル値を丸める
                                processedSample = this.roundToPrecision(processedSample);
                            }

                            outputChannel[i] = processedSample;
                        }

                    } else {
                         if (outputChannel) outputChannel.fill(0);
                    }
                }
            }
            return true;
        }
    }
    registerProcessor('local-enhance-processor', LocalEnhanceProcessor);
    `;

    const blobOptimized = new Blob([processorCodeOptimized], { type: 'application/javascript' });
    let moduleURLOptimized = URL.createObjectURL(blobOptimized);

    const activeAudioContexts = [];
    let realTimeAudioContext = null;
    let mediaSourceNode = null;
    let workletNode = null;


     // リアルタイムAudioContextをセットアップし、ノードグラフを構築・接続
    async function setupRealtimeAudio(media, fundamentalFrequency, physicalEvaluationResults) {
        console.log("Setting up realtime AudioContext and node graph...");

        if (realTimeAudioContext && realTimeAudioContext.state !== 'closed') {
            await realTimeAudioContext.close().catch(e => console.error("Failed to close existing AudioContext:", e));
            activeAudioContexts.splice(activeAudioContexts.indexOf(realTimeAudioContext), 1);
            realTimeAudioContext = null;
        }

        realTimeAudioContext = new (window.AudioContext || window.webkitAudioContext)({
             sampleRate: TARGET_SAMPLE_RATE
        });
        activeAudioContexts.push(realTimeAudioContext);
         console.log(`Created Realtime AudioContext with sample rate: ${realTimeAudioContext.sampleRate} Hz (Target: ${TARGET_SAMPLE_RATE} Hz).`);


        try {
            await realTimeAudioContext.audioWorklet.addModule(moduleURLOptimized);
            console.log("Optimized AudioWorklet module (Advanced Concepts) added.");
        } catch (e) {
            console.error("Failed to add Optimized AudioWorklet module:", e);
            throw new Error("Failed to add AudioWorklet module.");
        }

        mediaSourceNode = realTimeAudioContext.createMediaElementSource(media);
         console.log(`MediaElementSourceNode created. Channels: ${mediaSourceNode.channelCount}`);

        workletNode = new AudioWorkletNode(realTimeAudioContext, 'local-enhance-processor', {
            numberOfInputs: 1,
            numberOfOutputs: 1,
            outputChannelCount: [mediaSourceNode.channelCount]
        });
         console.log("AudioWorkletNode (local-enhance-processor) created.");

            // --- リアルタイムAudioContext ノードグラフ構築 (模擬IMAX処理) --- // コメント追加

        // MediaSource -> チャンネル独立処理 -> Worklet (LocalEnhance) -> 高度な空間・ダイナミクス処理 -> Destination
        let currentNode = mediaSourceNode;

         if (channelCount > 0) {
             // 各チャンネル独立処理の関数を呼び出し
             currentNode = createChannelIndependentProcessing(realTimeAudioContext, currentNode, channelCount);
             console.log("Connected MediaSource to ChannelIndependentProcessing.");


             // LocalEnhance Worklet を接続
             currentNode.connect(workletNode);
             currentNode = workletNode;
             console.log("Connected ChannelIndependentProcessing to Worklet.");


             // 高度な空間・ダイナミクス処理の関数を呼び出し
             currentNode = createAdvancedSpatialDynamicsProcessing(realTimeAudioContext, currentNode, channelCount, fundamentalFrequency);
             console.log("Connected Worklet to AdvancedSpatialDynamicsProcessing.");


             // 最終出力をDestinationに接続 (AdvancedSpatialDynamicsProcessing の最終出力は finalGainNode_rt)
             currentNode.connect(realTimeAudioContext.destination);

              console.log("Realtime AudioContext node graph connected to Destination.");

         } else {
             console.warn("Media source has zero channels. Connecting source directly to destination (no processing).");
             mediaSourceNode.connect(realTimeAudioContext.destination);
         }

              console.log("Realtime AudioContext node graph connected to Destination.");

         } else {
             console.warn("Media source has zero channels. Connecting source directly to destination (no processing).");
             mediaSourceNode.connect(realTimeAudioContext.destination);
         }


        // AudioContextの状態がsuspendedの場合はresumeを試みる
         if (realTimeAudioContext.state === 'suspended') {
             console.log("AudioContext is suspended. Attempting to resume...");
             try {
                await realTimeAudioContext.resume();
                console.log("AudioContext resumed.");
                // Contextがrunningになったらサブハーモニックオシレーターを開始
                 if (fundamentalFrequency > 0 && subOsc_rt && subOsc_rt.state === 'pending') {
                     subOsc_rt.start();
                      console.log("Subharmonic oscillator started after context resumed.");
                 }
             } catch (e) {
                 console.error("Failed to resume AudioContext (user interaction may be required):", e);
             }
         } else if (realTimeAudioContext.state === 'running') {
              if (fundamentalFrequency > 0 && subOsc_rt && subOsc_rt.state === 'pending') {
                   subOsc_rt.start();
                    console.log("Subharmonic oscillator started (context was already running).");
              }
         }

         console.log("Realtime audio setup complete.");
         console.log("If no sound is heard, try interacting with the page (e.g., click, play media) to resume the AudioContext.");
    }


    // AudioContextと関連リソースのクリーンアップ
    async function cleanupRealtimeAudio() {
        console.log("Cleaning up realtime AudioContext resources...");
        if (mediaSourceNode) { mediaSourceNode.disconnect(); mediaSourceNode = null; }
        if (workletNode) { workletNode.disconnect(); workletNode = null; }

        if (realTimeAudioContext && realTimeAudioContext.state !== 'closed') {
            await realTimeAudioContext.close().catch(e => console.error("Failed to close AudioContext during cleanup:", e));
             console.log("Realtime AudioContext closed.");
        }
        realTimeAudioContext = null;

        if (moduleURLOptimized) {
            URL.revokeObjectURL(moduleURLOptimized);
            moduleURLOptimized = null;
             console.log("AudioWorklet Blob URL revoked.");
        }
         console.log("Realtime audio cleanup complete.");
    }


    // --- リソース解放処理 ---
    window.addEventListener('unload', async () => {
        console.log("Page unloading. Performing cleanup.");
         await cleanupRealtimeAudio();

        activeAudioContexts.forEach(ctx => {
            if (ctx && ctx.state !== 'closed') {
                ctx.close().catch(() => {});
            }
        });
        activeAudioContexts.length = 0;
         console.log("Unload cleanup complete.");
    });

     // エラー発生時のクリーンアップ
     const cleanupOnError = async (e) => {
         console.error("An error occurred during execution:", e);
         await cleanupRealtimeAudio();

         activeAudioContexts.forEach(ctx => {
             if (ctx && ctx.state !== 'closed') {
                 ctx.close().catch(() => {});
             }
         });
         activeAudioContexts.length = 0;
         console.log("Error cleanup complete.");
     };


    // UserScript実行開始
    try {
        const media = await waitForMedia();
        await primeDecode(media, PRIME_DECODE_WAIT_SEC); // Autoplay policy対策

        // 物理モデルの評価計算 (一度実行)
         const physicalEvaluationResults = {};
         const evaluationFrequency = 1000;
         const sampleCountForEvaluation = 10;
         const sourceAmplitude_P0 = 1.0;
         physicalEvaluationResults.hornExitPressure = [];
         for (let i = 0; i <= sampleCountForEvaluation; i++) {
            const t = (i / sampleCountForEvaluation) * (1.0 / evaluationFrequency) * 0.1 ;
            const exact = pPlane(HORN_L, t, evaluationFrequency, sourceAmplitude_P0);
            const approx = pApprox(HORN_L, t, evaluationFrequency, sourceAmplitude_P0);
            physicalEvaluationResults.hornExitPressure.push({ t: t, exact: exact, approx: approx });
         }
         const evalDistance_x = HORN_L / 2;
         physicalEvaluationResults.impedanceAtMidHorn = Z(evalDistance_x);
         const evalFreqTube1 = TUBE_RESONANCE_FREQ * 0.8;
         const evalFreqTube2 = TUBE_RESONANCE_FREQ;
         const evalFreqTube3 = TUBE_RESONANCE_FREQ * 1.2;
         physicalEvaluationResults.tubeResonanceGain = {
             freq1: { freq: evalFreqTube1, gain: calculateTubeResonanceGain(evalFreqTube1, TUBE_RESONANCE_FREQ, C) },
             freq2: { freq: evalFreqTube2, gain: calculateTubeResonanceGain(evalFreqTube2, TUBE_RESONANCE_FREQ, C) },
             freq3: { freq: evalFreqTube3, gain: calculateTubeResonanceGain(evalFreqTube3, TUBE_RESONANCE_FREQ, C) }
         };
         console.log("Physical model evaluation completed (results stored in variable).");


        // 基本周波数推定のためのデータ収集と推定
        
        console.log("Starting f0 estimation process...");
        const tempCtxForF0 = new AudioContext();
        activeAudioContexts.push(tempCtxForF0);
        const tempSourceForF0 = tempCtxForF0.createMediaElementSource(media);
        const analyserNodeForF0 = tempCtxForF0.createAnalyser();
        analyserNodeForF0.fftSize = 2048;

        tempSourceForF0.connect(analyserNodeForF0);

         const wasPausedForF0 = media.paused;
         let playAttemptSuccessfulForF0 = false;
         if (wasPausedForF0) {
             try {
                 const playPromise = media.play();
                 if (playPromise !== undefined) {
                     await playPromise;
                 }
                 playAttemptSuccessfulForF0 = true;
                 console.log("Media play started for f0 estimation data tapping.");
             } catch (e) {
                  console.warn("Media play failed for f0 estimation (may require user interaction):", e);
             }
         } else {
               playAttemptSuccessfulForF0 = true;
               console.log("Media was already playing. Starting f0 estimation data tapping.");
         }

        await new Promise(r => setTimeout(r, TAP_AUDIO_DURATION_SEC * 1000));
        console.log(`Finished f0 estimation data tapping (${TAP_AUDIO_DURATION_SEC} seconds).`);

        const fundamentalFrequency = estimateFundamental(analyserNodeForF0);
        console.log(`Estimated f0: ${fundamentalFrequency.toFixed(2)} Hz`);

        tempSourceForF0.disconnect();
        analyserNodeForF0.disconnect();
        if (tempCtxForF0 && tempCtxForF0.state !== 'closed') {
             tempCtxForF0.close().catch(e => console.error("Failed to close temporary AudioContext for f0:", e));
        }
        activeAudioContexts.splice(activeAudioContexts.indexOf(tempCtxForF0), 1);
         console.log("Temporary AudioContext for f0 estimation closed.");

        if (wasPausedForF0 && playAttemptSuccessfulForF0) {
            media.pause();
            console.log("Media paused after f0 estimation data tapping.");
        } else if (!wasPausedForF0 && playAttemptSuccessfulForF0) {
             console.log("Media continued playing after f0 estimation data tapping.");
        } else if (!playAttemptSuccessfulForF0) {
              console.warn("Media play for f0 estimation failed, original state might be affected.");
        }


        // リアルタイムAudioContextをセットアップし、ノードグラフを構築・接続
        await setupRealtimeAudio(media, fundamentalFrequency, physicalEvaluationResults);

        console.log("Audio processing graph is ready. Control playback using the media element.");
        console.log("If no sound is heard, try interacting with the page (e.g., click, play media) to resume the AudioContext.");


    } catch (e) {
        // エラー発生時はクリーンアップ
        cleanupOnError(e);
    }

})();




